/**
 * runso - run shared object
 * (C) 2000 Copyright by Guido Draheim <guidod@gmx.de>
 */
#include <stdio.h>
#include <dlfcn.h>
#include <limits.h>
#include <string.h>
 
static const char helpstring[] =
    " %s [-Llibpath -llib]... libs_main [args]... \n"
    "   RUN SHARED OBJECT "VERSION" (C) LGPL 2000 Guido Draheim <guidod@gmx.de> \n"
    " the program will try to dlopen(3) the given libs (in the libpaths) and when \n"
    " successful it will lookup the given dlsym(libs_main) and call it just like \n"
    " a normal C program's main function. The libs are all optional. If the \n"
    " libs_main is not accessible after initially, an implicit -llibs is deduced \n"
    " from the given main-function by simple subtraction of _main from the end \n"
    " of the symbol (if named like that). \n"
    " Example: \n"
    "    'runso my_toast_main -D/dev/toaster' will open <libmy_toast.so> as found in \n"
    "    the standard libpath and call the symbol <my_toast_main> with argc/argv set to \n"
    "    argc= 2 argv[0]=\"toast_main\" argv[1]=\"-D/dev/toaster\" \n"
    " Options: \n"
    "    -L<path>       adds the given path to the search path of shared objects \n"
    "    -l<obj>        will dlopen the libobj.so as found in the search path \n"
    " BUGS: \n"
    "    -L isn't working yet :-( \n"
    ;
    
struct loadedlib
{
    void* lib;
    struct loadedlib* prev;
};

static const char dll_ext[] = DLL_EXT;
static char* programbasename = 0;
static struct loadedlib* lastloadedlib = 0;

/* helper function */
static void*
_mainlibopen(char* funcname)
{
    auto char libname[PATH_MAX];
    register void* lib;
    register int i = strlen(funcname);
    memcpy (libname, funcname, i+1);
    if (i > 6 && !memcmp("_main", libname+(i-5),5)) libname[i-5] = '\0';
    strcat (libname, dll_ext);

    if ((lib = dlopen(libname, RTLD_GLOBAL|RTLD_NOW)))
        return lib;
        
    if (i > 3 && !!memcmp("lib",libname,3) && i+3 < PATH_MAX)
    {	memmove(libname+3,libname,i);
        memcpy(libname,"lib",3);
        if ((lib = dlopen(libname, RTLD_GLOBAL|RTLD_NOW)))
            return lib;
    }

    if (strrchr(libname, '_')) 
    {   *(strrchr(libname, '_')) = '\0';
        strcat (libname, dll_ext);
    }
    
    if ((lib = dlopen(libname, RTLD_GLOBAL|RTLD_NOW)))
        return lib;
       		/* yes, if "simplename_main" the dlerror will be set on "libsimplename.so" now */

    return 0;
}

/* recursive function */
static int
_main(int argc, char** argv)
{
    if (!argc) 
    {   fprintf(stderr, helpstring, programbasename);
        exit (1);
    }

    if (strlen(argv[0]) > 2 && !memcmp( argv[0], "-l", 2))
    {   void* lib = dlopen (argv[0]+2, RTLD_GLOBAL|RTLD_NOW);
        if(!lib) 
        {   fprintf(stderr, "opening lib %s failed: %s \n", 
      		argv[0], dlerror());
            return (255);
        }else
        {   register int r;
            auto struct loadedlib loaded = { lib, lastloadedlib };
            lastloadedlib = &loaded;
            r = _main(argc-1, argv+1); /* recursion */
            dlclose(lib);
            return r;
        }
    } else
    {   register void* sym;
        {   register struct loadedlib* loaded = lastloadedlib;
            for (; loaded ; loaded = loaded->prev)
            {  if ((sym = dlsym(loaded->lib, argv[0])))
                return (((int (*)(int,char**))sym)(argc, argv));
            }
        }{  register void* lib = _mainlibopen(argv[0]);
            if (!lib)
            {   fprintf(stderr, "loading %s from its own lib failed: %s \n", argv[0], dlerror());
                return 255;
            }
            if ((sym = dlsym(lib, argv[0])))
            {   int r;
                r = (((int (*)(int, char**))sym)(argc, argv));
                dlclose(lib);
                return r;
            } else
            {   fprintf(stderr, "calling %s from its own lib failed: %s \n", argv[0], dlerror());
                dlclose(lib);
                return 255;
            }
        }
    }
}
   
int
main(int argc, char** argv)
{
    programbasename = argv[0];
    if (strchr(programbasename, '/'))
        programbasename = strrchr(programbasename, '/') + 1;
  
    return _main(argc-1, argv+1);
}

